import argparse
import time

from nucleus_driver._messages import Messages
from nucleus_driver._connection import Connection
from nucleus_driver._logger import Logger
from nucleus_driver._parser import Parser
from nucleus_driver._commands import Commands
from nucleus_driver._flash import Flash
from nucleus_driver._assert import Assert
from nucleus_driver._syslog import Syslog
from nucleus_driver._download import Download


class NucleusDriver:

    def __init__(self):

        self.messages = Messages()
        self.connection = Connection(messages=self.messages)
        self.logger = Logger(messages=self.messages, connection=self.connection)
        self.parser = Parser(messages=self.messages, logger=self.logger, connection=self.connection)
        self.commands = Commands(messages=self.messages, connection=self.connection, parser=self.parser)
        self.flash = Flash(messages=self.messages, connection=self.connection, commands=self.commands)
        self.asserts = Assert(messages=self.messages, connection=self.connection, commands=self.commands)
        self.syslog = Syslog(messages=self.messages, connection=self.connection, commands=self.commands)
        self.download = Download(messages=self.messages, connection=self.connection, commands=self.commands, parser=self.parser, logger=self.logger)

        self.connection.commands = self.commands
        self.logger.commands = self.commands

        self.logging_fieldcal = False

    ###########################################
    # Connection
    ###########################################

    def set_serial_configuration(self, port=None):

        self.connection.set_serial_configuration(port=port)

    def set_tcp_configuration(self, host=None):

        self.connection.set_tcp_configuration(host=host)

    def connect(self, connection_type) -> bool:

        CONNECTION_TYPES = ['serial', 'tcp']

        if connection_type not in CONNECTION_TYPES:
            self.messages.write_warning('Connection type {} not in {}'.format(connection_type, CONNECTION_TYPES))
            return False

        return self.connection.connect(connection_type=connection_type)

    def disconnect(self) -> bool:

        return self.connection.disconnect()

    ###########################################
    # Command
    ###########################################

    def send_command(self, command: str):

        command = command.rstrip('\n').rstrip('\r')

        if command.upper() == 'START':
            self.messages.write_message('start command is only supported through logging')

        elif command.upper() == 'FIELDCAL':

            self.messages.write_message('fieldcal command is only supported through logging')

        elif command.upper() == 'STARTSPECTRUM':

            self.messages.write_message('startspectrum command is only supported through logging')

        elif command.upper() == 'STOP':

            self.messages.write_message('stop command is only supported through logging')

        else:
            self.connection.write(command.encode() + b'\r\n')
            for i in range(100):
                command_reply = self.connection.readline()
                if command_reply == b'':
                    break
                if command_reply:
                    try:
                        self.messages.write_message(command_reply.decode())
                    except Exception as e:
                        self.messages.write_exception('Could not decode reply: {}'.format(command_reply))
                        self.messages.write_exception('Received error: {}'.format(e))

    ###########################################
    # Logging
    ###########################################

    def set_log_path(self, path):

        self.logger.set_path(path=path)

    def start_logging(self):

        self.logger.start()

    def stop_logging(self):

        self.logger.stop()

    ###########################################
    # Parser
    ###########################################

    def read_packet(self):

        packet = self.parser.read_packet()

        return packet

    def read_ascii(self):

        packet = self.parser.read_ascii()

        return packet

    def read_condition(self):

        packet = self.parser.read_condition()

        return packet

    ###########################################
    # Measurement
    ###########################################

    def start_measurement(self):

        self.logger.get_cp_nc()
        self.parser.start()
        self.commands._start()

    def start_fieldcal(self):

        self.parser.start()
        self.commands._fieldcal()

        self.logging_fieldcal = True

    def stop(self):

        self.parser.stop()
        if self.logging_fieldcal:
            time.sleep(0.5)
        self.commands._stop()

        self.logging_fieldcal = False

    ###########################################
    # Flash
    ###########################################

    def flash_firmware(self, path: str) -> bool:

        self.flash.reset_flash_files()

        if not self.flash.set_flash_files(path=path):
            self.messages.write_warning('Was not able to set flash file: {}'.format(path))
            return False

        result = self.flash.flash_firmware()

        if result == 0:
            self.messages.write_message('Successfully flashed firmware')
            return True
        else:
            self.messages.write_warning('Failed to flash firmware with error: {}'.format(result))
            return True

    ###########################################
    # Assert
    ###########################################

    def set_assert_path(self, path: str):

        return self.asserts.set_file_path(path=path)

    def clear_assert(self) -> [bytes]:

        return self.asserts.clear_assert()

    def assert_download(self):

        self.asserts.read_assert()

        status = False
        if len(self.asserts.assert_encrypted) > 1:
            self.asserts.write_encrypted_assert_to_file()
            status = True

        return status

    ###########################################
    # Syslog
    ###########################################

    def set_syslog_path(self, path: str):

        return self.syslog.set_file_path(path=path)

    def clear_syslog(self) -> [bytes]:

        return self.syslog.clear_syslog()

    def syslog_download(self):

        self.syslog.read_syslog()

        status = False
        if len(self.syslog.syslog_encrypted) > 1:
            self.syslog.write_encrypted_syslog_to_file()
            status = True

        return status

    ###########################################
    # Download
    ###########################################

    def set_download_path(self, path: str):

        return self.download.set_path(path=path)

    def download_dvl_data(self, fid=None, sa=None, length=None):

        return self.download.download_dvl_data(fid=fid, sa=sa, length=length)

    def download_nucleus_data(self, fid=None, sa=None, length=None):

        return self.download.download_nucleus_data(fid=fid, sa=sa, length=length)

    def convert_nucleus_data(self, path):

        return self.download.convert_nucleus_data(path=path)

    def list_files(self, src=None):

        return self.commands.list_files(src=src)



from queue import Queue, Empty
from struct import unpack
from itertools import zip_longest
from threading import Thread
from datetime import datetime
import time


class Parser:
    FAMILY_ID_NUCLEUS = 0x20
    FAMILY_ID_DVL = 0x16

    ID_IMU = 0x82
    ID_MAGNETOMETER = 0x87
    ID_BOTTOMTRACK = 0xb4
    ID_WATERTRACK = 0xbe
    ID_ALTIMETER = 0xaa
    ID_AHRS = 0xd2
    ID_INS = 0xdc
    ID_FIELD_CALIBRATION = 0x8B
    ID_ASCII = 0xA0
    ID_SPECTRUM_ANALYZER = 0x20
    ID_CURRENT_PROFILE = 0xC0

    DEFAULT_RESPONSE_TIMEOUT = 5

    MAX_PACKAGE_LENGTH = 7000

    # MAX_PACKAGE_LENGTH is given by the following firmware scripts:
    # https://dev.azure.com/NortekGroup/INS/_git/Firmware?path=/src/ofmt.c&version=GBmaster&line=28&lineEnd=29&lineStartColumn=1&lineEndColumn=1&lineStyle=plain&_a=contents
    # https://dev.azure.com/NortekGroup/INS/_git/Firmware?path=/lib/sensor_fusion/src/uns_types.h&version=GBmaster&line=10&lineEnd=11&lineStartColumn=1&lineEndColumn=1&lineStyle=plain&_a=contents

    def __init__(self, **kwargs):

        self.messages = kwargs.get('messages')
        self.logger = kwargs.get('logger')
        self.connection = kwargs.get('connection')

        self.packet_queue = Queue(maxsize=10000)
        self.ascii_queue = Queue(maxsize=10000)
        self.condition_queue = Queue(maxsize=10000)

        self.reading_packet = False
        self.binary_packet = bytearray()
        self.ascii_packet = list()

        self._queuing = {'packet': True,
                         'ascii': True,
                         'condition': True}

        self.thread = Thread()
        self.thread_running = False

    def init_packet_queue(self, maxsize=0):

        if not self.thread_running:
            self.packet_queue = Queue(maxsize=maxsize)
        else:
            self.messages.write_warning('can not initiate queue when parser is running')

    def init_condition_queue(self, maxsize=0):

        if not self.thread_running:
            self.condition_queue = Queue(maxsize=maxsize)
        else:
            self.messages.write_warning('can not initiate queue when parser is running')

    def init_ascii_queue(self, maxsize=0):

        if not self.thread_running:
            self.ascii_queue = Queue(maxsize=maxsize)
        else:
            self.messages.write_warning('can not initiate queue when parser is running')

    def set_queuing(self, packet: bool = None, ascii: bool = None, condition: bool = None):

        if packet is not None:
            self._queuing['packet'] = packet

        if ascii is not None:
            self._queuing['ascii'] = ascii

        if condition is not None:
            self._queuing['condition'] = condition

    def get_queuing(self):

        return self._queuing

    def clear_queue(self, queue_name):

        if queue_name == 'packet' or queue_name == 'all':
            for i in range(self.packet_queue.qsize()):
                self.packet_queue.get_nowait()

        if queue_name == 'condition' or queue_name == 'all':
            for i in range(self.condition_queue.qsize()):
                self.condition_queue.get_nowait()

        if queue_name == 'ascii' or queue_name == 'all':

            for i in range(self.ascii_queue.qsize()):
                self.ascii_queue.get_nowait()

    def write_packet(self, packet):

        if self._queuing['packet'] is True:
            self.packet_queue.put_nowait(packet)

        if self.logger._logging is True:

            self.logger._writing_packet = True

            try:
                self.logger.packet_writer.writerow(packet)
            except ValueError as exception:
                self.messages.write_warning('Failed to write package to csv file: {}'.format(exception))

            self.logger._writing_packet = False

    def read_packet(self, timeout=None):

        packet = None

        try:
            if timeout is not None:
                packet = self.packet_queue.get(timeout=timeout)

            elif not self.packet_queue.empty():
                packet = self.packet_queue.get_nowait()

        except Exception as exception:
            self.messages.write_warning('failed to retrieve packet from packet queue: {}'.format(exception))

        return packet

    def write_ascii(self, packet):

        if self._queuing['ascii'] is True:
            ascii_bytes = bytes(packet)
            ascii_packet = {'timestamp_python': datetime.now().timestamp(),
                            'bytes': ascii_bytes}

            self.ascii_queue.put_nowait(ascii_packet)

        if self.logger._logging is True:
            ascii_message = ''.join(chr(i) for i in packet if 0 <= i <= 0x7e).rstrip('\r\n')
            ascii_packet = {'timestamp_python': datetime.now().timestamp(),
                            'message': ascii_message}

            self.logger._writing_ascii = True

            try:
                self.logger.ascii_writer.writerow(ascii_packet)
            except ValueError as exception:
                self.messages.write_warning('Failed to write ascii message to csv file: {}'.format(exception))

            self.logger._writing_ascii = False

    def read_ascii(self, timeout=None):

        packet = None

        try:
            if timeout is not None:
                packet = self.ascii_queue.get(timeout=timeout)

            else:
                packet = self.ascii_queue.get(timeout=self.connection.timeout)

        except Empty:
            pass

        except Exception as exception:
            self.messages.write_warning('failed to retrieve packet from ascii queue: {}'.format(exception))

        return packet

    def write_condition(self, error_message, packet):

        failed_packet = {'timestamp_python': datetime.now().timestamp(),
                         'error_message': error_message,
                         'failed_packet': packet}

        if self._queuing['condition'] is True:
            self.condition_queue.put_nowait(failed_packet)

        if self.logger._logging is True:

            self.logger._writing_condition = True

            try:
                self.logger.condition_writer.writerow(failed_packet)
            except ValueError as exception:
                self.messages.write_warning('Failed to write condition to csv file: {}'.format(exception))

            self.logger._writing_condition = False

    def read_condition(self, timeout=None):

        packet = None

        try:
            if timeout is not None:
                packet = self.condition_queue.get(timeout=timeout)

            elif not self.condition_queue.empty():
                packet = self.condition_queue.get_nowait()

        except Exception as exception:
            self.messages.write_warning('failed to retrieve packet from condition queue: {}'.format(exception))

        return packet

    @staticmethod
    def checksum(packet):

        checksum = 0xb58c

        for u, v in zip_longest(packet[::2], packet[1::2], fillvalue=None):

            if v is not None:
                checksum += u | v << 8
            else:
                checksum += u << 8 | 0x00

            checksum &= 0xffff

        return checksum

    def add_binary_packet(self, binary_packet, ascii_packet):

        header_checksum, data_checksum, packet = self.get_packet(binary_packet)

        if header_checksum:
            if data_checksum and packet['id'] == hex(self.ID_ASCII):
                self.write_ascii(packet)

            elif data_checksum:
                self.write_packet(packet)

            else:
                self.write_condition(error_message='data checksum failed', packet=binary_packet)

            binary_packet = binary_packet[packet['sizeHeader'] + packet['sizeData']:]
            if len(binary_packet) == 0:
                reading_packet = False
            else:
                reading_packet = True

        elif 0xa5 in binary_packet[1:]:
            start_index = next(i + 1 for i, x in enumerate(binary_packet[1:]) if x == 0xa5)
            ascii_packet.extend(binary_packet[:start_index])
            binary_packet = binary_packet[start_index:]
            reading_packet = True

        else:
            self.write_condition(error_message='header checksum failed', packet=binary_packet)
            reading_packet = False
            binary_packet = bytearray()

        return binary_packet, ascii_packet, reading_packet

    def get_packet(self, binary_packet):

        header_checksum = False
        data_checksum = False
        packet = dict()

        if not isinstance(binary_packet, bytearray):
            self.messages.write_exception('packet is not bytearray. Extraction aborted')
            return header_checksum, data_checksum, packet

        def _get_status(status_bits, bit):

            status = False

            if (status_bits >> bit) & 0x01 == 1:
                status = True

            return status

        def check_header_size():

            status = False

            if len(binary_packet) >= unpack('<B', binary_packet[1:2])[0]:
                status = True

            return status

        def check_packet_size():

            status = False

            if len(binary_packet) >= header_data['sizeHeader'] + header_data['sizeData']:
                status = True

            return status

        def get_header_data():

            header = {'sizeHeader': unpack('<B', binary_packet[1:2])[0],
                      'id': unpack('<B', binary_packet[2:3])[0],
                      'family': unpack('<B', binary_packet[3:4])[0],
                      'sizeData': unpack('<H', binary_packet[4:6])[0],
                      'size': unpack('<B', binary_packet[1:2])[0] + unpack('<H', binary_packet[4:6])[0],
                      'dataCheckSum': unpack('<H', binary_packet[6:8])[0],
                      'headerCheckSum': unpack('<H', binary_packet[8:10])[0]}

            return header

        def get_common_data():

            status = unpack('<B', data[2:3])[0]

            common = {'version': unpack('<B', data[0:1])[0],
                      'offsetOfData': unpack('<B', data[1:2])[0],
                      'flags.posixTime': _get_status(status_bits=status, bit=0),
                      'timeStamp': unpack('<I', data[4:8])[0],
                      'microSeconds': unpack('<I', data[8:12])[0]}

            return common

        def get_sensor_data():

            sensor = None

            if header_data['family'] == self.FAMILY_ID_NUCLEUS:

                if header_data['id'] == self.ID_AHRS:

                    sensor = {'serialNumber': unpack('<I', data[16:20])[0],
                              'operationMode': unpack('<B', data[24:25])[0],  # TODO: add operationModeString?
                              'ahrsData.roll': unpack('<f', data[common_data['offsetOfData']: common_data['offsetOfData'] + 4])[0],
                              'ahrsData.pitch': unpack('<f', data[common_data['offsetOfData'] + 4: common_data['offsetOfData'] + 8])[0],
                              'ahrsData.heading': unpack('<f', data[common_data['offsetOfData'] + 8: common_data['offsetOfData'] + 12])[0],
                              'ahrsData.quaternionW': unpack('<f', data[common_data['offsetOfData'] + 12: common_data['offsetOfData'] + 16])[0],
                              'ahrsData.quaternionX': unpack('<f', data[common_data['offsetOfData'] + 16: common_data['offsetOfData'] + 20])[0],
                              'ahrsData.quaternionY': unpack('<f', data[common_data['offsetOfData'] + 20: common_data['offsetOfData'] + 24])[0],
                              'ahrsData.quaternionZ': unpack('<f', data[common_data['offsetOfData'] + 24: common_data['offsetOfData'] + 28])[0],
                              'ahrsData.rotationMatrix_0': unpack('<f', data[common_data['offsetOfData'] + 28: common_data['offsetOfData'] + 32])[0],
                              'ahrsData.rotationMatrix_1': unpack('<f', data[common_data['offsetOfData'] + 32: common_data['offsetOfData'] + 36])[0],
                              'ahrsData.rotationMatrix_2': unpack('<f', data[common_data['offsetOfData'] + 36: common_data['offsetOfData'] + 40])[0],
                              'ahrsData.rotationMatrix_3': unpack('<f', data[common_data['offsetOfData'] + 40: common_data['offsetOfData'] + 44])[0],
                              'ahrsData.rotationMatrix_4': unpack('<f', data[common_data['offsetOfData'] + 44: common_data['offsetOfData'] + 48])[0],
                              'ahrsData.rotationMatrix_5': unpack('<f', data[common_data['offsetOfData'] + 48: common_data['offsetOfData'] + 52])[0],
                              'ahrsData.rotationMatrix_6': unpack('<f', data[common_data['offsetOfData'] + 52: common_data['offsetOfData'] + 56])[0],
                              'ahrsData.rotationMatrix_7': unpack('<f', data[common_data['offsetOfData'] + 56: common_data['offsetOfData'] + 60])[0],
                              'ahrsData.rotationMatrix_8': unpack('<f', data[common_data['offsetOfData'] + 60: common_data['offsetOfData'] + 64])[0],
                              'declination': unpack('<f', data[common_data['offsetOfData'] + 64: common_data['offsetOfData'] + 68])[0],
                              'depth': unpack('<f', data[common_data['offsetOfData'] + 68: common_data['offsetOfData'] + 72])[0]
                              }

                    if common_data['version'] == 1:
                        sensor['fomAhrs'] = unpack('<B', data[25:26])[0]
                        sensor['fomFc1'] = unpack('<B', data[26:27])[0]

                    if common_data['version'] == 2:
                        sensor['fomAhrs'] = unpack('<f', data[28:32])[0]
                        sensor['fomFc1'] = unpack('<f', data[32:36])[0]

                if header_data['id'] == self.ID_INS:

                    sensor = {'serialNumber': unpack('<I', data[16:20])[0],
                              'operationMode': unpack('<B', data[24:25])[0],  # TODO: add operationModeString?
                              'ahrsData.roll': unpack('<f', data[common_data['offsetOfData']: common_data['offsetOfData'] + 4])[0],
                              'ahrsData.pitch': unpack('<f', data[common_data['offsetOfData'] + 4: common_data['offsetOfData'] + 8])[0],
                              'ahrsData.heading': unpack('<f', data[common_data['offsetOfData'] + 8: common_data['offsetOfData'] + 12])[0],
                              'ahrsData.quaternionW': unpack('<f', data[common_data['offsetOfData'] + 12: common_data['offsetOfData'] + 16])[0],
                              'ahrsData.quaternionX': unpack('<f', data[common_data['offsetOfData'] + 16: common_data['offsetOfData'] + 20])[0],
                              'ahrsData.quaternionY': unpack('<f', data[common_data['offsetOfData'] + 20: common_data['offsetOfData'] + 24])[0],
                              'ahrsData.quaternionZ': unpack('<f', data[common_data['offsetOfData'] + 24: common_data['offsetOfData'] + 28])[0],
                              'ahrsData.rotationMatrix_0': unpack('<f', data[common_data['offsetOfData'] + 28: common_data['offsetOfData'] + 32])[0],
                              'ahrsData.rotationMatrix_1': unpack('<f', data[common_data['offsetOfData'] + 32: common_data['offsetOfData'] + 36])[0],
                              'ahrsData.rotationMatrix_2': unpack('<f', data[common_data['offsetOfData'] + 36: common_data['offsetOfData'] + 40])[0],
                              'ahrsData.rotationMatrix_3': unpack('<f', data[common_data['offsetOfData'] + 40: common_data['offsetOfData'] + 44])[0],
                              'ahrsData.rotationMatrix_4': unpack('<f', data[common_data['offsetOfData'] + 44: common_data['offsetOfData'] + 48])[0],
                              'ahrsData.rotationMatrix_5': unpack('<f', data[common_data['offsetOfData'] + 48: common_data['offsetOfData'] + 52])[0],
                              'ahrsData.rotationMatrix_6': unpack('<f', data[common_data['offsetOfData'] + 52: common_data['offsetOfData'] + 56])[0],
                              'ahrsData.rotationMatrix_7': unpack('<f', data[common_data['offsetOfData'] + 56: common_data['offsetOfData'] + 60])[0],
                              'ahrsData.rotationMatrix_8': unpack('<f', data[common_data['offsetOfData'] + 60: common_data['offsetOfData'] + 64])[0],
                              'declination': unpack('<f', data[common_data['offsetOfData'] + 64: common_data['offsetOfData'] + 68])[0],
                              'depth': unpack('<f', data[common_data['offsetOfData'] + 68: common_data['offsetOfData'] + 72])[0],
                              'deltaQuaternionW': unpack('<f', data[common_data['offsetOfData'] + 72: common_data['offsetOfData'] + 76])[0],
                              'deltaQuaternionX': unpack('<f', data[common_data['offsetOfData'] + 76: common_data['offsetOfData'] + 80])[0],
                              'deltaQuaternionY': unpack('<f', data[common_data['offsetOfData'] + 80: common_data['offsetOfData'] + 84])[0],
                              'deltaQuaternionZ': unpack('<f', data[common_data['offsetOfData'] + 84: common_data['offsetOfData'] + 88])[0],
                              'courseOverGround': unpack('<f', data[common_data['offsetOfData'] + 88: common_data['offsetOfData'] + 92])[0],
                              'temperature': unpack('<f', data[common_data['offsetOfData'] + 92: common_data['offsetOfData'] + 96])[0],
                              'pressure': unpack('<f', data[common_data['offsetOfData'] + 96: common_data['offsetOfData'] + 100])[0],
                              'altitude': unpack('<f', data[common_data['offsetOfData'] + 100: common_data['offsetOfData'] + 104])[0],
                              'latitude': unpack('<f', data[common_data['offsetOfData'] + 104: common_data['offsetOfData'] + 108])[0],
                              'longitude': unpack('<f', data[common_data['offsetOfData'] + 108: common_data['offsetOfData'] + 112])[0],
                              'height': unpack('<f', data[common_data['offsetOfData'] + 112: common_data['offsetOfData'] + 116])[0],
                              'positionFrameX': unpack('<f', data[common_data['offsetOfData'] + 116: common_data['offsetOfData'] + 120])[0],
                              'positionFrameY': unpack('<f', data[common_data['offsetOfData'] + 120: common_data['offsetOfData'] + 124])[0],
                              'positionFrameZ': unpack('<f', data[common_data['offsetOfData'] + 124: common_data['offsetOfData'] + 128])[0],
                              'deltaPositionFrameX': unpack('<f', data[common_data['offsetOfData'] + 128: common_data['offsetOfData'] + 132])[0],
                              'deltaPositionFrameY': unpack('<f', data[common_data['offsetOfData'] + 132: common_data['offsetOfData'] + 136])[0],
                              'deltaPositionFrameZ': unpack('<f', data[common_data['offsetOfData'] + 136: common_data['offsetOfData'] + 140])[0],
                              'deltaPositionNucleusX': unpack('<f', data[common_data['offsetOfData'] + 140: common_data['offsetOfData'] + 144])[0],
                              'deltaPositionNucleusY': unpack('<f', data[common_data['offsetOfData'] + 144: common_data['offsetOfData'] + 148])[0],
                              'deltaPositionNucleusZ': unpack('<f', data[common_data['offsetOfData'] + 148: common_data['offsetOfData'] + 152])[0],
                              'velocityNedX': unpack('<f', data[common_data['offsetOfData'] + 152: common_data['offsetOfData'] + 156])[0],
                              'velocityNedY': unpack('<f', data[common_data['offsetOfData'] + 156: common_data['offsetOfData'] + 160])[0],
                              'velocityNedZ': unpack('<f', data[common_data['offsetOfData'] + 160: common_data['offsetOfData'] + 164])[0],
                              'velocityNucleusX': unpack('<f', data[common_data['offsetOfData'] + 164: common_data['offsetOfData'] + 168])[0],
                              'velocityNucleusY': unpack('<f', data[common_data['offsetOfData'] + 168: common_data['offsetOfData'] + 172])[0],
                              'velocityNucleusZ': unpack('<f', data[common_data['offsetOfData'] + 172: common_data['offsetOfData'] + 176])[0],
                              'deltaVelocityNedX': unpack('<f', data[common_data['offsetOfData'] + 176: common_data['offsetOfData'] + 180])[0],
                              'deltaVelocityNedY': unpack('<f', data[common_data['offsetOfData'] + 180: common_data['offsetOfData'] + 184])[0],
                              'deltaVelocityNedZ': unpack('<f', data[common_data['offsetOfData'] + 184: common_data['offsetOfData'] + 188])[0],
                              'deltaVelocityNucleusX': unpack('<f', data[common_data['offsetOfData'] + 188: common_data['offsetOfData'] + 192])[0],
                              'deltaVelocityNucleusY': unpack('<f', data[common_data['offsetOfData'] + 192: common_data['offsetOfData'] + 196])[0],
                              'deltaVelocityNucleusZ': unpack('<f', data[common_data['offsetOfData'] + 196: common_data['offsetOfData'] + 200])[0],
                              'speedOverGround': unpack('<f', data[common_data['offsetOfData'] + 200: common_data['offsetOfData'] + 204])[0],
                              'turnRateX': unpack('<f', data[common_data['offsetOfData'] + 204: common_data['offsetOfData'] + 208])[0],
                              'turnRateY': unpack('<f', data[common_data['offsetOfData'] + 208: common_data['offsetOfData'] + 212])[0],
                              'turnRateZ': unpack('<f', data[common_data['offsetOfData'] + 212: common_data['offsetOfData'] + 216])[0],
                              }

                    if common_data['version'] == 1:
                        sensor['fomAhrs'] = unpack('<B', data[25:26])[0]
                        sensor['fomFc1'] = unpack('<B', data[26:27])[0]

                    if common_data['version'] == 2:
                        sensor['fomAhrs'] = unpack('<f', data[28:32])[0]
                        sensor['fomFc1'] = unpack('<f', data[32:36])[0]

                if header_data['id'] == self.ID_IMU:
                    imu_status = unpack('<I', data[12:16])[0]

                    sensor = {'status.isValid': _get_status(status_bits=imu_status, bit=0),
                              'status.hasDataPathOverrun': _get_status(status_bits=imu_status, bit=17),
                              'status.hasFlashUpdateFailure': _get_status(status_bits=imu_status, bit=18),
                              'status.hasSpiComError': _get_status(status_bits=imu_status, bit=19),
                              'status.hasLowVoltage': _get_status(status_bits=imu_status, bit=20),
                              'status.hasSensorFailure': _get_status(status_bits=imu_status, bit=21),
                              'status.hasMemoryFailure': _get_status(status_bits=imu_status, bit=22),
                              'status.hasGyro1Failure': _get_status(status_bits=imu_status, bit=23),
                              'status.hasGyro2Failure': _get_status(status_bits=imu_status, bit=24),
                              'status.hasAccelerometerFailure': _get_status(status_bits=imu_status, bit=25),
                              'accelerometer.x': unpack('<f', data[common_data['offsetOfData']: common_data['offsetOfData'] + 4])[0],
                              'accelerometer.y': unpack('<f', data[common_data['offsetOfData'] + 4: common_data['offsetOfData'] + 8])[0],
                              'accelerometer.z': unpack('<f', data[common_data['offsetOfData'] + 8: common_data['offsetOfData'] + 12])[0],
                              'gyro.x': unpack('<f', data[common_data['offsetOfData'] + 12: common_data['offsetOfData'] + 16])[0],
                              'gyro.y': unpack('<f', data[common_data['offsetOfData'] + 16: common_data['offsetOfData'] + 20])[0],
                              'gyro.z': unpack('<f', data[common_data['offsetOfData'] + 20: common_data['offsetOfData'] + 24])[0],
                              'temperature': unpack('<f', data[common_data['offsetOfData'] + 24: common_data['offsetOfData'] + 28])[0]
                              }

                if header_data['id'] == self.ID_MAGNETOMETER:
                    mag_status = unpack('<I', data[12:16])[0]

                    sensor = {'status.isCompensatedForHardIron': _get_status(status_bits=mag_status, bit=0),
                              'status.dvlActive': _get_status(status_bits=mag_status, bit=29),
                              'status.dvlAcousticsActive': _get_status(status_bits=mag_status, bit=30),
                              'status.dvlTransmitterActive': _get_status(status_bits=mag_status, bit=31),
                              'magnetometer.x': unpack('<f', data[common_data['offsetOfData']: common_data['offsetOfData'] + 4])[0],
                              'magnetometer.y': unpack('<f', data[common_data['offsetOfData'] + 4: common_data['offsetOfData'] + 8])[0],
                              'magnetometer.z': unpack('<f', data[common_data['offsetOfData'] + 8: common_data['offsetOfData'] + 12])[0]
                              }

                if header_data['id'] in (self.ID_BOTTOMTRACK, self.ID_WATERTRACK):
                    status = unpack('<I', data[12:16])[0]

                    sensor = {'status.beam1VelocityValid': _get_status(status_bits=status, bit=0),
                              'status.beam2VelocityValid': _get_status(status_bits=status, bit=1),
                              'status.beam3VelocityValid': _get_status(status_bits=status, bit=2),
                              'status.beam1DistanceValid': _get_status(status_bits=status, bit=3),
                              'status.beam2DistanceValid': _get_status(status_bits=status, bit=4),
                              'status.beam3DistanceValid': _get_status(status_bits=status, bit=5),
                              'status.beam1FomValid': _get_status(status_bits=status, bit=6),
                              'status.beam2FomValid': _get_status(status_bits=status, bit=7),
                              'status.beam3FomValid': _get_status(status_bits=status, bit=8),
                              'status.xVelocityValid': _get_status(status_bits=status, bit=9),
                              'status.yVelocityValid': _get_status(status_bits=status, bit=10),
                              'status.zVelocityValid': _get_status(status_bits=status, bit=11),
                              'status.xFomValid': _get_status(status_bits=status, bit=12),
                              'status.yFomValid': _get_status(status_bits=status, bit=13),
                              'status.zFomValid': _get_status(status_bits=status, bit=14),
                              'serialNumber': unpack('<I', data[16:20])[0],
                              'soundSpeed': unpack('<f', data[24:28])[0],
                              'temperature': unpack('<f', data[28:32])[0],
                              'pressure': unpack('<f', data[32:36])[0],
                              'velocityBeam1': unpack('<f', data[36:40])[0],
                              'velocityBeam2': unpack('<f', data[40:44])[0],
                              'velocityBeam3': unpack('<f', data[44:48])[0],
                              'distanceBeam1': unpack('<f', data[48:52])[0],
                              'distanceBeam2': unpack('<f', data[52:56])[0],
                              'distanceBeam3': unpack('<f', data[56:60])[0],
                              'fomBeam1': unpack('<f', data[60:64])[0],
                              'fomBeam2': unpack('<f', data[64:68])[0],
                              'fomBeam3': unpack('<f', data[68:72])[0],
                              'dtBeam1': unpack('<f', data[72:76])[0],
                              'dtBeam2': unpack('<f', data[76:80])[0],
                              'dtBeam3': unpack('<f', data[80:84])[0],
                              'timeVelBeam1': unpack('<f', data[84:88])[0],
                              'timeVelBeam2': unpack('<f', data[88:92])[0],
                              'timeVelBeam3': unpack('<f', data[92:96])[0],
                              'velocityX': unpack('<f', data[96:100])[0],
                              'velocityY': unpack('<f', data[100:104])[0],
                              'velocityZ': unpack('<f', data[104:108])[0],
                              'fomX': unpack('<f', data[108:112])[0],
                              'fomY': unpack('<f', data[112:116])[0],
                              'fomZ': unpack('<f', data[116:120])[0],
                              'dtXYZ': unpack('<f', data[120:124])[0],
                              'timeVelXYZ': unpack('<f', data[124:128])[0]
                              }

                if header_data['id'] == self.ID_ALTIMETER:
                    status = unpack('<I', data[12:16])[0]

                    sensor = {'status.altimeterDistanceValid': _get_status(status_bits=status, bit=0),
                              'status.altimeterQualityValid': _get_status(status_bits=status, bit=1),
                              'status.pressureValid': _get_status(status_bits=status, bit=16),
                              'status.temperatureValid': _get_status(status_bits=status, bit=17),
                              'serialNumber': unpack('<I', data[16:20])[0],
                              'soundSpeed': unpack('<f', data[24:28])[0],
                              'temperature': unpack('<f', data[28:32])[0],
                              'pressure': unpack('<f', data[32:36])[0],
                              'altimeterDistance': unpack('<f', data[36:40])[0],
                              'altimeterQuality': unpack('<H', data[40:42])[0]
                              }

                if header_data['id'] == self.ID_CURRENT_PROFILE:

                    sensor = {'serialNumber': unpack('<I', data[16:20])[0],
                              'soundVelocity': unpack('<f', data[24:28])[0],
                              'temperature': unpack('<f', data[28:32])[0],
                              'pressure': unpack('<f', data[32:36])[0],
                              'cellSize': unpack('<f', data[36:40])[0],
                              'blanking': unpack('<f', data[40:44])[0],
                              'numberOfCells': unpack('<H', data[44:46])[0],
                              'ambiguityVelocity': unpack('<H', data[46:48])[0]}

                    velocity_data_format = '<' + ''.ljust(3 * sensor['numberOfCells'], 'H')
                    velocity_data_offset = common_data['offsetOfData']
                    velocity_data_size = 2 * 3 * sensor['numberOfCells']
                    # sensor['velocityData'] = unpack(velocity_data_format, data[velocity_data_offset: velocity_data_offset + velocity_data_size])[:velocity_data_size]
                    velocity_data = unpack(velocity_data_format, data[velocity_data_offset: velocity_data_offset + velocity_data_size])[:velocity_data_size]
                    for index, velocity_cell in enumerate(velocity_data):
                        sensor['velocityData_{}'.format(index)] = velocity_cell

                    amplitude_data_format = '<' + ''.ljust(3 * sensor['numberOfCells'], 'B')
                    amplitude_data_offset = common_data['offsetOfData'] + 6 * sensor['numberOfCells']
                    amplitude_data_size = 1 * 3 * sensor['numberOfCells']
                    # sensor['amplitudeData'] = unpack(amplitude_data_format, data[amplitude_data_offset: amplitude_data_offset + amplitude_data_size])[:amplitude_data_size]
                    amplitude_data = unpack(amplitude_data_format, data[amplitude_data_offset: amplitude_data_offset + amplitude_data_size])[:amplitude_data_size]
                    for index, amplitude_cell in enumerate(amplitude_data):
                        sensor['amplitudeData_{}'.format(index)] = amplitude_cell

                    correlation_data_format = '<' + ''.ljust(3 * sensor['numberOfCells'], 'B')
                    correlation_data_offset = common_data['offsetOfData'] + 9 * sensor['numberOfCells']
                    correlation_data_size = 1 * 3 * sensor['numberOfCells']
                    # sensor['correlationData'] = unpack(correlation_data_format, data[correlation_data_offset: correlation_data_offset + correlation_data_size])[:correlation_data_size]
                    correlation_data = unpack(correlation_data_format, data[correlation_data_offset: correlation_data_offset + correlation_data_size])[:correlation_data_size]
                    for index, correlation_cell in enumerate(correlation_data):
                        sensor['correlationData_{}'.format(index)] = correlation_cell

                if header_data['id'] == self.ID_FIELD_CALIBRATION:
                    status = unpack('<I', data[12:16])[0]

                    sensor = {'status.pointsUsedInEstimation': _get_status(status_bits=status, bit=0),
                              'hardIron.x': unpack('<f', data[common_data['offsetOfData']: common_data['offsetOfData'] + 4])[0],
                              'hardIron.y': unpack('<f', data[common_data['offsetOfData'] + 4: common_data['offsetOfData'] + 8])[0],
                              'hardIron.z': unpack('<f', data[common_data['offsetOfData'] + 8: common_data['offsetOfData'] + 12])[0],
                              'sAxis_0': unpack('<f', data[common_data['offsetOfData'] + 12: common_data['offsetOfData'] + 16])[0],
                              'sAxis_1': unpack('<f', data[common_data['offsetOfData'] + 16: common_data['offsetOfData'] + 20])[0],
                              'sAxis_2': unpack('<f', data[common_data['offsetOfData'] + 20: common_data['offsetOfData'] + 24])[0],
                              'sAxis_3': unpack('<f', data[common_data['offsetOfData'] + 24: common_data['offsetOfData'] + 28])[0],
                              'sAxis_4': unpack('<f', data[common_data['offsetOfData'] + 28: common_data['offsetOfData'] + 32])[0],
                              'sAxis_5': unpack('<f', data[common_data['offsetOfData'] + 32: common_data['offsetOfData'] + 36])[0],
                              'sAxis_6': unpack('<f', data[common_data['offsetOfData'] + 36: common_data['offsetOfData'] + 40])[0],
                              'sAxis_7': unpack('<f', data[common_data['offsetOfData'] + 40: common_data['offsetOfData'] + 44])[0],
                              'sAxis_8': unpack('<f', data[common_data['offsetOfData'] + 44: common_data['offsetOfData'] + 48])[0],
                              'newPoint.x': unpack('<f', data[common_data['offsetOfData'] + 48: common_data['offsetOfData'] + 52])[0],
                              'newPoint.y': unpack('<f', data[common_data['offsetOfData'] + 52: common_data['offsetOfData'] + 56])[0],
                              'newPoint.z': unpack('<f', data[common_data['offsetOfData'] + 56: common_data['offsetOfData'] + 60])[0],
                              'fomFieldCalibration': unpack('<f', data[common_data['offsetOfData'] + 60: common_data['offsetOfData'] + 64])[0],
                              'coverage': unpack('<f', data[common_data['offsetOfData'] + 64: common_data['offsetOfData'] + 68])[0]
                              }

                if header_data['id'] == self.ID_ASCII:
                    sensor = {'string': unpack('<{}s'.format(len(data)), data)}

            if header_data['family'] == self.FAMILY_ID_DVL:

                if header_data['id'] == self.ID_SPECTRUM_ANALYZER:
                    sensor = {'data': data}

            return sensor

        if not check_header_size():
            self.messages.write_exception('Packet is smaller than specified header length. Extraction aborted')
            return header_checksum, data_checksum, packet

        header_data = get_header_data()

        if self.checksum(binary_packet[:header_data['sizeHeader'] - 2]) == header_data['headerCheckSum']:
            header_checksum = True
        else:
            self.messages.write_exception('Header did not pass checksum. Extraction aborted')
            return header_checksum, data_checksum, packet

        packet.update(header_data)

        if not check_packet_size():
            self.messages.write_exception('Packet is smaller than specified header and data length. Extraction aborted')
            return header_checksum, data_checksum, packet

        if self.checksum(binary_packet[header_data['sizeHeader']: header_data['sizeHeader'] + header_data['sizeData']]) == header_data['dataCheckSum']:
            data_checksum = True
        else:
            self.messages.write_exception('Packet did not pass checksum. Extraction aborted')
            return header_checksum, data_checksum, packet

        data = binary_packet[header_data['sizeHeader']:header_data['sizeHeader'] + header_data['sizeData']]

        common_data = get_common_data()
        packet.update(common_data)

        packet['timestampPython'] = datetime.now().timestamp()

        sensor_data = get_sensor_data()

        if sensor_data is None:
            self.messages.write_exception('Unable to unpack sensor data. Extraction aborted')
            return header_checksum, data_checksum, packet

        packet.update(sensor_data)

        return header_checksum, data_checksum, packet

    def add_ascii_packet(self, ascii_packet):

        self.write_ascii(packet=ascii_packet)

    def add_data(self, data):
        for value in data:
            if value == 0xa5 and not self.reading_packet:
                self.reading_packet = True
                self.ascii_packet = list()

            if self.reading_packet:
                self.binary_packet.append(value)
            else:
                self.ascii_packet.append(value)

            if len(self.binary_packet) > self.MAX_PACKAGE_LENGTH or len(self.binary_packet) > 5 and len(self.binary_packet) >= self.binary_packet[1] + (self.binary_packet[4] & 0xff) | ((self.binary_packet[5] & 0xff) << 8):
                self.binary_packet, self.ascii_packet, self.reading_packet = self.add_binary_packet(self.binary_packet, self.ascii_packet)

            if len(self.ascii_packet) >= 2 and self.ascii_packet[-2] == 0x0d and self.ascii_packet[-1] == 0x0a:
                self.add_ascii_packet(self.ascii_packet)
                self.ascii_packet = list()

    def start(self) -> bool:

        if self.thread.is_alive():
            self.messages.write_warning(message='Nucleus thread is already alive')
            return False

        self.thread = Thread(target=self.run)
        self.thread.start()

        return True

    def stop(self) -> bool:

        if not self.thread.is_alive():
            self.messages.write_warning(message='Nucleus thread is not alive')
            return False

        self.thread_running = False
        self.thread.join(2)
        self.thread = Thread()

        return True

    def run(self):

        self.thread_running = True

        while self.thread_running:

            if not self.connection.get_connection_status():
                time.sleep(0.05)
                continue

            data = self.connection.read()

            if data:

                self.add_data(data=data)

            else:
                time.sleep(0.01)



from re import match
from datetime import datetime


class Commands:
    DHCP_STATIC = ['DHCP', 'STATIC']
    ADDRESS_PATTERN = r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$'

    def __init__(self, **kwargs):

        self.connection = kwargs.get('connection')
        self.messages = kwargs.get('messages')
        self.parser = kwargs.get('parser')

    def _reset_buffer(self):

        if self.parser.thread_running is not True:
            self.connection.reset_buffers()
        elif self.parser.get_queuing()['ascii'] is True:
            self.parser.clear_queue(queue_name='ascii')

    def _get_reply(self, terminator: bytes = None, timeout: int = 1) -> [bytes]:

        get_reply = b''

        if self.parser.thread_running is not True:
            get_reply = self.connection.read(terminator=terminator, timeout=timeout)
        elif self.parser.get_queuing()['ascii'] is True:
            init_time = datetime.now()
            while (datetime.now() - init_time).seconds < timeout:
                ascii_packet = self.parser.read_ascii()
                if ascii_packet is not None:
                    get_reply += ascii_packet['bytes']

                if terminator is not None and terminator in get_reply:
                    break

        return get_reply

    def _check_reply(self, data: bytes, command: bytes, terminator: bytes = None):

        if terminator is not None and terminator not in data:
            if b'ERROR' in data:
                get_error_reply = self.get_error().rstrip(b'OK\r\n')
                self.messages.write_exception(message='Received ERROR instead of {} after sending {}: {}'.format(terminator, command, get_error_reply))

            else:
                self.messages.write_warning(message='Did not receive {} after sending {}: {}'.format(terminator, command, data))

        elif terminator is None:
            if b'ERROR' in data:
                get_error_reply = self.get_error().rstrip(b'OK\r\n')
                self.messages.write_exception(message='Received ERROR after sending {}: {}'.format(command, get_error_reply))

    def _handle_reply(self, command, terminator: bytes = None, timeout: int = 1, nmea=False) -> [bytes]:

        if nmea:
            terminator = b'$PNOR,' + terminator.rstrip(b'\r\n')
            nmea_checksum = self._nmea_checksum(terminator)
            terminator = terminator + b'*' + nmea_checksum + b'\r\n'

        get_reply = self._get_reply(terminator=terminator, timeout=timeout)
        self._check_reply(data=get_reply, terminator=terminator, command=command)

        reply_list = [i + b'\r\n' for i in get_reply.split(b'\r\n') if i]

        if nmea:
            for reply in reply_list:
                reply_split = reply.split(b'*')
                nmea_checksum = self._nmea_checksum(reply_split[0])
                if nmea_checksum != reply_split[1].rstrip(b'\r\n'):
                    self.messages.write_warning('Reply did not pass nmea checksum: {}\t{}'.format(reply, nmea_checksum))

        return reply_list

    def _nmea_checksum(self, command):

        checksum = 0

        command = command.lstrip(b'$')

        for byte in command:
            checksum ^= byte

        return '{0:02X}'.format(checksum).encode()

    def get_error(self) -> [bytes]:

        self._reset_buffer()

        command = b'GETERROR\r\n'

        self.connection.write(command)

        get_reply = self._get_reply(terminator=b'OK\r\n')

        return get_reply

    def read_start_up(self, timeout: int = 1) -> [bytes]:

        reply = self._get_reply(terminator=b'OK\r\n', timeout=timeout)
        # print(reply)
        # for i in range(5):
        #    reply += self._get_reply()
        #    print(reply)
        #    if b'OK\r\n' in reply:
        #        break

        # TODO: Add _handle_reply() ?

        start_up_message = [reply]

        return start_up_message

    def _start(self) -> [bytes]:

        self._reset_buffer()

        command = b'START\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n', timeout=2)

        return get_reply

    def _stop(self, timeout=1) -> [bytes]:

        self._reset_buffer()

        command = b'STOP\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n', timeout=timeout)

        return get_reply

    def trig(self) -> [bytes]:

        self._reset_buffer()

        trig_command = b'TRIG\r\n'

        self.connection.write(trig_command)

        get_reply = self._handle_reply(command=trig_command, terminator=b'OK\r\n')

        return get_reply

    def _fieldcal(self) -> [bytes]:

        self._reset_buffer()

        fieldcal_command = b'FIELDCAL\r\n'

        self.connection.write(fieldcal_command)

        get_reply = self._handle_reply(command=fieldcal_command, terminator=b'OK\r\n')

        return get_reply

    def reboot(self) -> [bytes]:

        self._reset_buffer()

        command = b'REBOOT\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    # TODO: in default driver?
    def _fw_update(self, timeout=38) -> [bytes]:

        self._reset_buffer()

        command = b'FWUPDATE\r\n'

        self.connection.write(command)
        get_ack_reply = self._handle_reply(command=command, terminator=b'ACK\r\n')

        if b'ACK\r\n' not in get_ack_reply:
            return get_ack_reply

        # The following timeout is defined based on the maximum time each of the following operations may take.
        # The timeouts are given by the following calculations: https://dev.azure.com/NortekGroup/9f822050-97e3-47d7-9697-3773608ceeff/_apis/git/repositories/4dea5944-f474-4ef9-b642-a82f0e306b66/items?path=/doc/flash.ods&versionDescriptor%5BversionOptions%5D=0&versionDescriptor%5BversionType%5D=0&versionDescriptor%5Bversion%5D=master&resolveLfs=true&%24format=octetStream&api-version=5.0&download=true
        get_ok_reply = self._handle_reply(command=command, terminator=b'OK\r\n', timeout=timeout)

        get_reply = list()
        get_reply.extend(get_ack_reply)
        get_reply.extend(get_ok_reply)

        return get_reply

    # TODO: in default driver?
    def _fw_confirm(self):

        self._reset_buffer()

        command = b'FWCONFIRM\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    # TODO: in default driver?
    def _dvl_update(self) -> [bytes]:

        self._reset_buffer()

        command = b'DVLUPDATE\r\n'

        self.connection.write(command)
        get_reply = self._handle_reply(command=command, terminator=b'ACK\r\n')

        return get_reply

    # TODO: in default driver?
    def _upload(self, package) -> [bytes]:

        self._reset_buffer()

        command = b'UPLOAD\r\n'

        self.connection.write(command)
        get_ack_reply = self._handle_reply(command=command, terminator=b'ACK\r\n')

        if b'ACK\r\n' not in get_ack_reply:
            return get_ack_reply

        self.connection.write(package)
        get_ok_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        get_reply = list()
        get_reply.extend(get_ack_reply)
        get_reply.extend(get_ok_reply)

        return get_reply

    def save_all(self) -> [bytes]:

        self._reset_buffer()

        command = b'SAVE,ALL\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def save_config(self) -> [bytes]:

        self._reset_buffer()

        save_config_command = b'SAVE,CONFIG\r\n'

        self.connection.write(save_config_command)

        get_reply = self._handle_reply(command=save_config_command, terminator=b'OK\r\n')

        return get_reply

    def save_mission(self) -> [bytes]:

        self._reset_buffer()

        save_mission_command = b'SAVE,MISSION\r\n'

        self.connection.write(save_mission_command)

        get_reply = self._handle_reply(command=save_mission_command, terminator=b'OK\r\n')

        return get_reply

    def save_magcal(self) -> [bytes]:

        self._reset_buffer()

        save_magcal_command = b'SAVE,MAGCAL\r\n'

        self.connection.write(save_magcal_command)

        get_reply = self._handle_reply(command=save_magcal_command, terminator=b'OK\r\n')

        return get_reply

    def save_comm(self) -> [bytes]:

        self._reset_buffer()

        command = b'SAVE,COMM\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def set_default_all(self) -> [bytes]:

        self._reset_buffer()

        command = b'SETDEFAULT,ALL\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def set_default_config(self) -> [bytes]:

        self._reset_buffer()

        set_default_config_command = b'SETDEFAULT,CONFIG\r\n'

        self.connection.write(set_default_config_command)

        get_reply = self._handle_reply(command=set_default_config_command, terminator=b'OK\r\n')

        return get_reply

    def set_default_mission(self) -> [bytes]:

        self._reset_buffer()

        set_default_mission_command = b'SETDEFAULT,MISSION\r\n'

        self.connection.write(set_default_mission_command)

        get_reply = self._handle_reply(command=set_default_mission_command, terminator=b'OK\r\n')

        return get_reply

    def set_default_magcal(self) -> [bytes]:

        self._reset_buffer()

        set_default_magcal_command = b'SETDEFAULT,MAGCAL\r\n'

        self.connection.write(set_default_magcal_command)

        get_reply = self._handle_reply(command=set_default_magcal_command, terminator=b'OK\r\n')

        return get_reply

    def set_default_comm(self) -> [bytes]:

        self._reset_buffer()

        command = b'SETDEFAULT,COMM\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def restore_all(self) -> [bytes]:

        self._reset_buffer()

        command = b'RESTORE,ALL\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def restore_config(self) -> [bytes]:

        self._reset_buffer()

        command = b'RESTORE,CONFIG\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def restore_mission(self) -> [bytes]:

        self._reset_buffer()

        command = b'RESTORE,MISSION\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def restore_magcal(self) -> [bytes]:

        self._reset_buffer()

        command = b'RESTORE,MAGCAL\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def restore_comm(self) -> [bytes]:

        self._reset_buffer()

        command = b'RESTORE,COMM\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def set_mission(self, poff=None, long=None, lat=None, decl=None, range=None, bd=None, sv=None,
                    sa=None) -> [bytes]:

        self._reset_buffer()

        set_mission_command = b'SETMISSION'

        if poff is not None:
            if isinstance(poff, float) or isinstance(poff, int):
                set_mission_command += b',POFF=' + str(poff).encode()
            else:
                self.messages.write_warning('Invalid value for POFF in SETMISSION command')

        if long is not None:
            if isinstance(long, float) or isinstance(long, int):
                set_mission_command += b',LONG=' + str(long).encode()
            else:
                self.messages.write_warning('Invalid value for LONG in SETMISSION command')

        if lat is not None:
            if isinstance(lat, float) or isinstance(lat, int):
                set_mission_command += b',LAT=' + str(lat).encode()
            else:
                self.messages.write_warning('Invalid value for LAT in SETMISSION command')

        if decl is not None:
            if isinstance(decl, float) or isinstance(decl, int):
                set_mission_command += b',DECL=' + str(decl).encode()
            else:
                self.messages.write_warning('Invalid value for DECL in SETMISSION command')

        if range is not None:
            if isinstance(range, float) or isinstance(range, int):
                set_mission_command += b',RANGE=' + str(range).encode()
            else:
                self.messages.write_warning('Invalid value for RANGE in SETMISSION command')

        if bd is not None:
            if isinstance(bd, float) or isinstance(bd, int):
                set_mission_command += b',BD=' + str(bd).encode()
            else:
                self.messages.write_warning('Invalid value for BD in SETMISSION command')

        if sv is not None:
            if isinstance(sv, float) or isinstance(sv, int):
                set_mission_command += b',SV=' + str(sv).encode()
            else:
                self.messages.write_warning('Invalid value for SV in SETMISSION command')

        if sa is not None:
            if isinstance(sa, float) or isinstance(sa, int):
                set_mission_command += b',SA=' + str(sa).encode()
            else:
                self.messages.write_warning('Invalid value for SA in SETMISSION command')

        set_mission_command += b'\r\n'

        self.connection.write(set_mission_command)

        get_reply = self._handle_reply(command=set_mission_command, terminator=b'OK\r\n')

        return get_reply

    def get_mission(self, poff=False, long=False, lat=False, decl=False, range=False, bd=False, sv=False,
                    sa=False) -> [bytes]:

        self._reset_buffer()

        get_mission_command = b'GETMISSION'

        if poff is True:
            get_mission_command += b',POFF'

        if long is True:
            get_mission_command += b',LONG'

        if lat is True:
            get_mission_command += b',LAT'

        if decl is True:
            get_mission_command += b',DECL'

        if range is True:
            get_mission_command += b',RANGE'

        if bd is True:
            get_mission_command += b',BD'

        if sv is True:
            get_mission_command += b',SV'

        if sa is True:
            get_mission_command += b',SA'

        get_mission_command += b'\r\n'

        self.connection.write(get_mission_command)

        get_reply = self._handle_reply(command=get_mission_command, terminator=b'OK\r\n')

        return get_reply

    def get_mission_lim(self, poff=False, long=False, lat=False, decl=False, range=False, bd=False, sv=False,
                        sa=False) -> [bytes]:

        self._reset_buffer()

        get_mission_lim_command = b'GETMISSIONLIM'

        if poff is True:
            get_mission_lim_command += b',POFF'

        if long is True:
            get_mission_lim_command += b',LONG'

        if lat is True:
            get_mission_lim_command += b',LAT'

        if decl is True:
            get_mission_lim_command += b',DECL'

        if range is True:
            get_mission_lim_command += b',RANGE'

        if bd is True:
            get_mission_lim_command += b',BD'

        if sv is True:
            get_mission_lim_command += b',SV'

        if sa is True:
            get_mission_lim_command += b',SA'

        get_mission_lim_command += b'\r\n'

        self.connection.write(get_mission_lim_command)

        get_reply = self._handle_reply(command=get_mission_lim_command, terminator=b'OK\r\n')

        return get_reply

    def set_inst(self, type=None, rotxy=None, rotyz=None, rotxz=None) -> [bytes]:

        self._reset_buffer()

        set_inst_command = b'SETINST'

        if type is not None:
            # TODO: Implement the type parameter
            pass

        if rotxy is not None:
            if isinstance(rotxy, float) or isinstance(rotxy, int):
                set_inst_command += b',ROTXY=' + str(rotxy).encode()
            else:
                self.messages.write_warning('Invalid value for ROTXY in SETINST command')

        if rotyz is not None:
            if isinstance(rotyz, float) or isinstance(rotyz, int):
                set_inst_command += b',ROTYZ=' + str(rotyz).encode()
            else:
                self.messages.write_warning('Invalid value for ROTYZ in SETINST command')

        if rotxz is not None:
            if isinstance(rotxz, float) or isinstance(rotxz, int):
                set_inst_command += b',ROTXZ=' + str(rotxz).encode()
            else:
                self.messages.write_warning('Invalid value for ROTXZ in SETINST command')

        set_inst_command += b'\r\n'

        self.connection.write(set_inst_command)

        get_reply = self._handle_reply(command=set_inst_command, terminator=b'OK\r\n')

        return get_reply

    def get_inst(self, type=False, rotxy=False, rotyz=False, rotxz=False) -> [bytes]:

        self._reset_buffer()

        get_inst_command = b'GETINST'

        if type is True:
            get_inst_command += b',TYPE'

        if rotxy is True:
            get_inst_command += b',ROTXY'

        if rotyz is True:
            get_inst_command += b',ROTYZ'

        if rotxz is True:
            get_inst_command += b',ROTXZ'

        get_inst_command += b'\r\n'

        self.connection.write(get_inst_command)

        get_reply = self._handle_reply(command=get_inst_command, terminator=b'OK\r\n')

        return get_reply

    def get_inst_lim(self, type=False, rotxy=False, rotyz=False, rotxz=False) -> [bytes]:

        self._reset_buffer()

        get_inst_lim_command = b'GETINSTLIM'

        if type is True:
            get_inst_lim_command += b',TYPE'

        if rotxy is True:
            get_inst_lim_command += b',ROTXY'

        if rotyz is True:
            get_inst_lim_command += b',ROTYZ'

        if rotxz is True:
            get_inst_lim_command += b',ROTXZ'

        get_inst_lim_command += b'\r\n'

        self.connection.write(get_inst_lim_command)

        get_reply = self._handle_reply(command=get_inst_lim_command, terminator=b'OK\r\n')

        return get_reply

    def set_ahrs(self, freq=None, mode=None, ds=None, df=None) -> [bytes]:

        self._reset_buffer()

        set_ahrs_command = b'SETAHRS'

        if freq is not None:
            if isinstance(freq, int):
                set_ahrs_command += b',FREQ=' + str(freq).encode()
            else:
                self.messages.write_warning('Invalid value for FREQ in SETAHRS command')

        if mode is not None:
            if isinstance(mode, int):
                set_ahrs_command += b',MODE=' + str(mode).encode()
            else:
                self.messages.write_warning('Invalid value for MODE in SETAHRS command')

        if ds is not None:
            if ds.upper() in self.ON_OFF:
                set_ahrs_command += b',DS="' + ds.upper().encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for DS in SETAHRS command')

        if df is not None:
            if isinstance(df, int):
                set_ahrs_command += b',DF=' + str(df).encode()
            else:
                self.messages.write_warning('Invalid value for DF in SETAHRS command')

        set_ahrs_command += b'\r\n'

        self.connection.write(set_ahrs_command)

        get_reply = self._handle_reply(command=set_ahrs_command, terminator=b'OK\r\n')

        return get_reply

    def get_ahrs(self, freq=False, mode=False, ds=False, df=False) -> [bytes]:

        self._reset_buffer()

        get_ahrs_command = b'GETAHRS'

        if freq is True:
            get_ahrs_command += b',FREQ'

        if mode is True:
            get_ahrs_command += b',MODE'

        if ds is True:
            get_ahrs_command += b',DS'

        if df is True:
            get_ahrs_command += b',DF'

        get_ahrs_command += b'\r\n'

        self.connection.write(get_ahrs_command)

        get_reply = self._handle_reply(command=get_ahrs_command, terminator=b'OK\r\n')

        return get_reply

    def get_ahrs_lim(self, freq=False, mode=False, ds=False, df=False) -> [bytes]:

        self._reset_buffer()

        get_ahrs_lim_command = b'GETAHRSLIM'

        if freq is True:
            get_ahrs_lim_command += b',FREQ'

        if mode is True:
            get_ahrs_lim_command += b',MODE'

        if ds is True:
            get_ahrs_lim_command += b',DS'

        if df is True:
            get_ahrs_lim_command += b',DF'

        get_ahrs_lim_command += b'\r\n'

        self.connection.write(get_ahrs_lim_command)

        get_reply = self._handle_reply(command=get_ahrs_lim_command, terminator=b'OK\r\n')

        return get_reply

    def set_fieldcal(self, mode=None) -> [bytes]:

        self._reset_buffer()

        command = b'SETFIELDCAL'

        if mode is not None:
            if isinstance(mode, int) and mode in [1, 2]:
                command += b',MODE=' + str(mode).encode()
            else:
                self.messages.write_warning('Invalid value for MODE in SETFIELDCAL command')

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_fieldcal(self, mode=False) -> [bytes]:

        self._reset_buffer()

        command = b'GETFIELDCAL'

        if mode is True:
            command += b',MODE'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_fieldcal_lim(self, mode=False) -> [bytes]:

        self._reset_buffer()

        command = b'GETFIELDCALLIM'

        if mode is True:
            command += b',MODE'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def set_bt(self, mode=None, vr=None, wt=None, pl=None, pl_mode=None, ds=None, df=None) -> [bytes]:

        self._reset_buffer()

        set_bt_command = b'SETBT'

        if mode is not None:
            if isinstance(mode, str) and mode.upper() in self.MODES:
                set_bt_command += b',MODE="' + mode.upper().encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for MODE in SETBT command')

        if vr is not None:
            if isinstance(vr, int) or isinstance(vr, float):
                set_bt_command += b',VR=' + str(vr).encode()
            else:
                self.messages.write_warning('Invalid value for VR in SETBT command')

        if wt is not None:
            if wt.upper() in self.ON_OFF:
                set_bt_command += b',WT="' + wt.upper().encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for WT in SETBT command')

        if pl is not None:
            if isinstance(pl, int) or isinstance(pl, float):
                set_bt_command += b',PL=' + str(pl).encode()
            else:
                self.messages.write_warning('Invalid value for PL in SETBT command')

        if pl_mode is not None:
            if isinstance(pl_mode, str) and pl_mode.upper() in self.PL_MODES:
                set_bt_command += b',PLMODE="' + pl_mode.upper().encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for PLMODE in SETBT command')

        if ds is not None:
            if ds.upper() in self.ON_OFF:
                set_bt_command += b',DS="' + ds.upper().encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for DS in SETBT command')

        if df is not None:
            if isinstance(df, int):
                set_bt_command += b',DF=' + str(df).encode()
            else:
                self.messages.write_warning('Invalid value for DF in SETBT command')

        set_bt_command += b'\r\n'

        self.connection.write(set_bt_command)

        get_reply = self._handle_reply(command=set_bt_command, terminator=b'OK\r\n')

        return get_reply

    def get_bt(self, vr=False, wt=False, pl=False, pl_mode=False, ds=False, df=False) -> [bytes]:

        self._reset_buffer()

        get_bt_command = b'GETBT'

        if vr is True:
            get_bt_command += b',VR'

        if wt is True:
            get_bt_command += b',WT'

        if pl is True:
            get_bt_command += b',PL'

        if pl_mode is True:
            get_bt_command += b',PLMODE'

        if ds is True:
            get_bt_command += b',DS'

        if df is True:
            get_bt_command += b',DF'

        get_bt_command += b'\r\n'

        self.connection.write(get_bt_command)

        get_reply = self._handle_reply(command=get_bt_command, terminator=b'OK\r\n')

        return get_reply

    def get_bt_lim(self, vr=False, wt=False, pl=False, pl_mode=False, ds=False, df=False) -> [bytes]:

        self._reset_buffer()

        get_bt_lim_command = b'GETBTLIM'

        if vr is True:
            get_bt_lim_command += b',VR'

        if wt is True:
            get_bt_lim_command += b',WT'

        if pl is True:
            get_bt_lim_command += b',PL'

        if pl_mode is True:
            get_bt_lim_command += b',PLMODE'

        if ds is True:
            get_bt_lim_command += b',DS'

        if df is True:
            get_bt_lim_command += b',DF'

        get_bt_lim_command += b'\r\n'

        self.connection.write(get_bt_lim_command)

        get_reply = self._handle_reply(command=get_bt_lim_command, terminator=b'OK\r\n')

        return get_reply

    def set_alti(self, pl=None, ds=None, df=None) -> [bytes]:

        self._reset_buffer()

        set_alti_command = b'SETALTI'

        if pl is not None:
            if isinstance(pl, int) or isinstance(pl, float):
                set_alti_command += b',PL=' + str(pl).encode()
            else:
                self.messages.write_warning('Invalid value for PL in SETALTI command')

        if ds is not None:
            if ds.upper() in self.ON_OFF:
                set_alti_command += b',DS="' + ds.upper().encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for DS in SETALTI command')

        if df is not None:
            if isinstance(df, int):
                set_alti_command += b',DF=' + str(df).encode()
            else:
                self.messages.write_warning('Invalid value for DF in SETALTI command')

        set_alti_command += b'\r\n'

        self.connection.write(set_alti_command)

        get_reply = self._handle_reply(command=set_alti_command, terminator=b'OK\r\n')

        return get_reply

    def get_alti(self, pl=False, ds=False, df=False) -> [bytes]:

        self._reset_buffer()

        get_alti_command = b'GETALTI'

        if pl is True:
            get_alti_command += b',PL'

        if ds is True:
            get_alti_command += b',DS'

        if df is True:
            get_alti_command += b',DF'

        get_alti_command += b'\r\n'

        self.connection.write(get_alti_command)

        get_reply = self._handle_reply(command=get_alti_command, terminator=b'OK\r\n')

        return get_reply

    def get_alti_lim(self, pl=False, ds=False, df=False) -> [bytes]:

        self._reset_buffer()

        get_alti_lim_command = b'GETALTILIM'

        if pl is True:
            get_alti_lim_command += b',PL'

        if ds is True:
            get_alti_lim_command += b',DS'

        if df is True:
            get_alti_lim_command += b',DF'

        get_alti_lim_command += b'\r\n'

        self.connection.write(get_alti_lim_command)

        get_reply = self._handle_reply(command=get_alti_lim_command, terminator=b'OK\r\n')

        return get_reply

    def set_cur_prof(self, profile_range=None, cs=None, bd=None, ds=None, df=None) -> [bytes]:

        self._reset_buffer()

        command = b'SETCURPROF'

        if profile_range is not None:
            if isinstance(profile_range, int) or isinstance(profile_range, float):
                command += b',RANGE=' + str(profile_range).encode()
            else:
                self.messages.write_warning('Invalid value for RANGE in SETCURPROF command')

        if cs is not None:
            if isinstance(cs, int) or isinstance(cs, float):
                command += b',CS=' + str(cs).encode()
            else:
                self.messages.write_warning('Invalid value for CS in SETCURPROF command')

        if bd is not None:
            if isinstance(bd, int) or isinstance(bd, float):
                command += b',BD=' + str(bd).encode()
            else:
                self.messages.write_warning('Invalid value for BD in SETCURPROF command')

        if ds is not None:
            if ds.upper() in self.ON_OFF:
                command += b',DS="' + ds.upper().encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for DS in SETCURPROF command')

        if df is not None:
            if isinstance(df, int):
                command += b',DF=' + str(df).encode()
            else:
                self.messages.write_warning('Invalid value for DF in SETCURPROF command')

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_cur_prof(self, profile_range=False, cs=False, bd=False, ds=False, df=False):

        self._reset_buffer()

        command = b'GETCURPROF'

        if profile_range is True:
            command += b',RANGE'

        if cs is True:
            command += b',CS'

        if bd is True:
            command += b',BD'

        if ds is True:
            command += b',DS'

        if df is True:
            command += b',DF'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_cur_prof_lim(self, profile_range=False, cs=False, bd=False, ds=False, df=False):

        self._reset_buffer()

        command = b'GETCURPROFLIM'

        if profile_range is True:
            command += b',RANGE'

        if cs is True:
            command += b',CS'

        if bd is True:
            command += b',BD'

        if ds is True:
            command += b',DS'

        if df is True:
            command += b',DF'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def set_trig(self, src=None, freq=None, alti=None, cp=None) -> [bytes]:

        self._reset_buffer()

        set_trig_command = b'SETTRIG'

        if src is not None:
            if src.upper() in self.TRIGGER_SOURCES:
                set_trig_command += b',SRC="' + src.upper().encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for SRC in SETTRIG command')

        if freq is not None:
            if isinstance(freq, int):
                set_trig_command += b',FREQ=' + str(freq).encode()
            else:
                self.messages.write_warning('Invalid value for FREQ in SETTRIG command')

        if alti is not None:
            if isinstance(alti, int) and alti in self.ALTI_RANGE:
                set_trig_command += b',ALTI=' + str(alti).encode()
            else:
                self.messages.write_warning('Invalid value for ALTI in SETTRIG command')

        if cp is not None:
            if isinstance(cp, int) and cp in self.CP_RANGE:
                set_trig_command += b',CP=' + str(cp).encode()
            else:
                self.messages.write_warning('Invalid value for CP in SETTRIG command')

        set_trig_command += b'\r\n'

        self.connection.write(set_trig_command)

        get_reply = self._handle_reply(command=set_trig_command, terminator=b'OK\r\n')

        return get_reply

    def get_trig(self, src=False, freq=False, alti=False, cp=False) -> [bytes]:

        self._reset_buffer()

        command = b'GETTRIG'

        if src is True:
            command += b',SRC'

        if freq is True:
            command += b',FREQ'

        if alti is True:
            command += b',ALTI'

        if cp is True:
            command += b',CP'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_trig_lim(self, src=False, freq=False, alti=False, cp=False) -> [bytes]:

        self._reset_buffer()

        command = b'GETTRIGLIM'

        if src is True:
            command += b',SRC'

        if freq is True:
            command += b',FREQ'

        if alti is True:
            command += b',ALTI'

        if cp is True:
            command += b',CP'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def set_mag(self, method=None) -> [bytes]:

        self._reset_buffer()

        command = b'SETMAG'

        if method is not None:
            if isinstance(method, str) and method.upper() in self.MAG_METHODS:
                command += b',METHOD="' + method.upper().encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for METHOD in SETMAG command')

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_mag(self, method=False) -> [bytes]:

        self._reset_buffer()

        command = b'GETMAG'

        if method is True:
            command += b',METHOD'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_mag_lim(self, method=False) -> [bytes]:

        self._reset_buffer()

        command = b'GETMAGLIM'

        if method is True:
            command += b',METHOD'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def set_mag_cal(self, hx=None, hy=None, hz=None, m11=None, m12=None, m13=None, m21=None, m22=None, m23=None,
                    m31=None, m32=None, m33=None) -> [bytes]:

        self._reset_buffer()

        set_mag_cal_command = b'SETMAGCAL'

        if hx is not None:
            if isinstance(hx, int) or isinstance(hx, float):
                set_mag_cal_command += b',HX=' + str(hx).encode()
            else:
                self.messages.write_warning('Invalid value for HX in SETMAGCAL command')

        if hy is not None:
            if isinstance(hy, int) or isinstance(hy, float):
                set_mag_cal_command += b',HY=' + str(hy).encode()
            else:
                self.messages.write_warning('Invalid value for HY in SETMAGCAL command')

        if hz is not None:
            if isinstance(hz, int) or isinstance(hz, float):
                set_mag_cal_command += b',HZ=' + str(hz).encode()
            else:
                self.messages.write_warning('Invalid value for HZ in SETMAGCAL command')

        if m11 is not None:
            if isinstance(m11, int) or isinstance(m11, float):
                set_mag_cal_command += b',M11=' + str(m11).encode()
            else:
                self.messages.write_warning('Invalid value for M11 in SETMAGCAL command')

        if m12 is not None:
            if isinstance(m12, int) or isinstance(m12, float):
                set_mag_cal_command += b',M12=' + str(m12).encode()
            else:
                self.messages.write_warning('Invalid value for M12 in SETMAGCAL command')

        if m13 is not None:
            if isinstance(m13, int) or isinstance(m13, float):
                set_mag_cal_command += b',M13=' + str(m13).encode()
            else:
                self.messages.write_warning('Invalid value for M13 in SETMAGCAL command')

        if m21 is not None:
            if isinstance(m21, int) or isinstance(m21, float):
                set_mag_cal_command += b',M21=' + str(m21).encode()
            else:
                self.messages.write_warning('Invalid value for M21 in SETMAGCAL command')

        if m22 is not None:
            if isinstance(m22, int) or isinstance(m22, float):
                set_mag_cal_command += b',M22=' + str(m22).encode()
            else:
                self.messages.write_warning('Invalid value for M22 in SETMAGCAL command')

        if m23 is not None:
            if isinstance(m23, int) or isinstance(m23, float):
                set_mag_cal_command += b',M23=' + str(m23).encode()
            else:
                self.messages.write_warning('Invalid value for M23 in SETMAGCAL command')

        if m31 is not None:
            if isinstance(m31, int) or isinstance(m31, float):
                set_mag_cal_command += b',M31=' + str(m31).encode()
            else:
                self.messages.write_warning('Invalid value for M31 in SETMAGCAL command')

        if m32 is not None:
            if isinstance(m32, int) or isinstance(m32, float):
                set_mag_cal_command += b',M32=' + str(m32).encode()
            else:
                self.messages.write_warning('Invalid value for M32 in SETMAGCAL command')

        if m33 is not None:
            if isinstance(m33, int) or isinstance(m33, float):
                set_mag_cal_command += b',M33=' + str(m33).encode()
            else:
                self.messages.write_warning('Invalid value for M33 in SETMAGCAL command')

        set_mag_cal_command += b'\r\n'

        self.connection.write(set_mag_cal_command)

        get_reply = self._handle_reply(command=set_mag_cal_command, terminator=b'OK\r\n')

        return get_reply

    def get_mag_cal(self, hx=False, hy=False, hz=False, m11=False, m12=False, m13=False, m21=False, m22=False,
                    m23=False, m31=False, m32=False, m33=False) -> [bytes]:

        self._reset_buffer()

        get_mag_cal_command = b'GETMAGCAL'

        if hx is True:
            get_mag_cal_command += b',HX'

        if hy is True:
            get_mag_cal_command += b',HY'

        if hz is True:
            get_mag_cal_command += b',HZ'

        if m11 is True:
            get_mag_cal_command += b',M11'

        if m12 is True:
            get_mag_cal_command += b',M12'

        if m13 is True:
            get_mag_cal_command += b',M13'

        if m21 is True:
            get_mag_cal_command += b',M21'

        if m22 is True:
            get_mag_cal_command += b',M22'

        if m23 is True:
            get_mag_cal_command += b',M23'

        if m31 is True:
            get_mag_cal_command += b',M31'

        if m32 is True:
            get_mag_cal_command += b',M32'

        if m33 is True:
            get_mag_cal_command += b',M33'

        get_mag_cal_command += b'\r\n'

        self.connection.write(get_mag_cal_command)

        get_reply = self._handle_reply(command=get_mag_cal_command, terminator=b'OK\r\n')

        return get_reply

    def get_mag_cal_lim(self, hx=False, hy=False, hz=False, m11=False, m12=False, m13=False, m21=False,
                        m22=False, m23=False, m31=False, m32=False, m33=False) -> [bytes]:

        self._reset_buffer()

        get_mag_cal_lim_command = b'GETMAGCAL'

        if hx is True:
            get_mag_cal_lim_command += b',HX'

        if hy is True:
            get_mag_cal_lim_command += b',HY'

        if hz is True:
            get_mag_cal_lim_command += b',HZ'

        if m11 is True:
            get_mag_cal_lim_command += b',M11'

        if m12 is True:
            get_mag_cal_lim_command += b',M12'

        if m13 is True:
            get_mag_cal_lim_command += b',M13'

        if m21 is True:
            get_mag_cal_lim_command += b',M21'

        if m22 is True:
            get_mag_cal_lim_command += b',M22'

        if m23 is True:
            get_mag_cal_lim_command += b',M23'

        if m31 is True:
            get_mag_cal_lim_command += b',M31'

        if m32 is True:
            get_mag_cal_lim_command += b',M32'

        if m33 is True:
            get_mag_cal_lim_command += b',M33'

        get_mag_cal_lim_command += b'\r\n'

        self.connection.write(get_mag_cal_lim_command)

        get_reply = self._handle_reply(command=get_mag_cal_lim_command, terminator=b'OK\r\n')

        return get_reply

    def set_eth(self, ipmethod=None, ip=None, netmask=None, gateway=None):

        self._reset_buffer()

        command = b'SETETH'

        if ipmethod is not None:
            if ipmethod.upper() in self.DHCP_STATIC:
                command += b',IPMETHOD="' + ipmethod.upper().encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for IPMETHOD in SETETH command')

        if ip is not None:
            if match(self.ADDRESS_PATTERN, ip):
                command += b',IP="' + ip.encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for IP in SETETH command')

        if netmask is not None:
            if match(self.ADDRESS_PATTERN, netmask):
                command += b',NETMASK="' + netmask.encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for NETMASK in SETETH command')

        if gateway is not None:
            if match(self.ADDRESS_PATTERN, gateway):
                command += b',GATEWAY="' + gateway.encode() + b'"'
            else:
                self.messages.write_warning('Invalid value for GATEWAY in SETETH command')

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_eth(self, ipmethod=False, ip=False, netmask=False, gateway=False):

        self._reset_buffer()

        command = b'GETETH'

        if ipmethod is True:
            command += b',IPMETHOD'

        if ip is True:
            command += b',IP'

        if netmask is True:
            command += b',NETMASK'

        if gateway is True:
            command += b',GATEWAY'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_eth_lim(self, ipmethod=False, ip=False, netmask=False, gateway=False):

        self._reset_buffer()

        command = b'GETETHLIM'

        if ipmethod is True:
            command += b',IPMETHOD'

        if ip is True:
            command += b',IP'

        if netmask is True:
            command += b',NETMASK'

        if gateway is True:
            command += b',GATEWAY'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def read_ip(self, ip=False, netmask=False, gateway=False, leasetime=False):

        self._reset_buffer()

        command = b'READIP'

        if ip is True:
            command += b',IP'

        if netmask is True:
            command += b',NETMASK'

        if gateway is True:
            command += b',GATEWAY'

        if leasetime is True:
            command += b',LEASETIME'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_id(self, sn=False, str=False) -> [bytes]:

        self._reset_buffer()

        command = b'ID'

        if sn is True:
            command += b',SN'

        if str is True:
            command += b',STR'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_hw(self, digital=False, analog=False) -> [bytes]:

        self._reset_buffer()

        command = b'GETHW'

        if digital is True:
            command += b',DIGITAL'

        if analog is True:
            command += b',ANALOG'

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_fw(self, str=False, major=False, minor=False, patch=False, extra=False, build=False, hash=False, dvlfw=False, dvlminor=False, dvlboot=False, dvlfpga=False, _nmea=False) -> [bytes]:

        self._reset_buffer()

        command = b''

        if _nmea is True:
            command += b'$PNOR,'

        command += b'GETFW'

        if str is True:
            command += b',STR'

        if major is True:
            command += b',MAJOR'

        if minor is True:
            command += b',MINOR'

        if patch is True:
            command += b',PATCH'

        if extra is True:
            command += b',EXTRA'

        if build is True:
            command += b',BUILD'

        if hash is True:
            command += b',HASH'

        if dvlfw is True:
            command += b',DVLFW'

        if dvlminor is True:
            command += b',DVLMINOR'

        if dvlboot is True:
            command += b',DVLBOOT'

        if dvlfpga is True:
            command += b',DVLFPGA'

        if _nmea is True:
            nmea_checksum = self._nmea_checksum(command)
            command += b'*'
            command += nmea_checksum

        #if _nmea is True:
        #    command = _append_nmea()

        command += b'\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n', nmea=_nmea)

        return get_reply

    def get_all(self):

        self._reset_buffer()

        command = b'GETALL\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def set_clockstring(self, clockstring: str) -> [bytes]:

        self._reset_buffer()

        command = 'SETCLOCKSTR,TIME="{}"\r\n'.format(clockstring).encode()

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def get_clockstring(self) -> [bytes]:

        self._reset_buffer()

        command = b'GETCLOCKSTR\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def list_license(self) -> [bytes]:

        self._reset_buffer()

        command = b'LISTLICENSE\r\n'

        self.connection.write(command)

        get_reply = self._handle_reply(command=command, terminator=b'OK\r\n')

        return get_reply

    def read_assert(self, timeout=1) -> [bytes]:

        self._reset_buffer()

        command = b'READASSERT\r\n'

        self.connection.write(command)

        reply = self._handle_reply(command=command, terminator=b'OK\r\n', timeout=timeout)

        get_reply = b''
        for entry in reply:
            get_reply += entry

        if get_reply[-4:] == b'OK\r\n':
            get_reply_list = [get_reply[:-4], get_reply[-4:]]
        else:
            get_reply_list = [get_reply]

        return get_reply_list

    def clear_assert(self) -> [bytes]:

        self._reset_buffer()

        set_default_mission_command = b'CLEARASSERT\r\n'

        self.connection.write(set_default_mission_command)

        get_reply = self._handle_reply(command=set_default_mission_command, terminator=b'OK\r\n')

        return get_reply

    def read_syslog(self) -> [bytes]:

        self._reset_buffer()

        read_syslog_command = b'READSYSLOG\r\n'

        self.connection.write(read_syslog_command)

        data = bytes()

        while True:

            readout = self.connection.read(terminator=b'OK\r\n', timeout=1)

            if readout == b'':
                break
            else:
                data += readout

        if data[-4:] == b'OK\r\n':
            data_list = [data[:-4], data[-4:]]
        else:
            data_list = [data]

        return data_list

    def clear_syslog(self) -> [bytes]:

        self._reset_buffer()

        clear_syslog_command = b'CLEARSYSLOG\r\n'

        self.connection.write(clear_syslog_command)

        get_reply = self._handle_reply(command=clear_syslog_command, terminator=b'OK\r\n', timeout=5)

        return get_reply

    def list_syslog(self) -> [bytes]:

        self._reset_buffer()

        list_syslog_command = b'LISTSYSLOG\r\n'

        self.connection.write(list_syslog_command)

        get_reply = self._handle_reply(command=list_syslog_command, terminator=b'OK\r\n')

        return get_reply

    def list_files(self, src=None):

        self._reset_buffer()

        listfiles_command = b'LISTFILES'

        if src is not None:
            if isinstance(src, int):
                listfiles_command += b',SRC=' + str(src).encode()
            else:
                self.messages.write_warning('Invalid value for SRC in DOWNLOAD command')

        listfiles_command += b'\r\n'

        self.connection.write(listfiles_command)

        get_reply = self._handle_reply(command=listfiles_command, terminator=b'OK\r\n', timeout=10)

        return get_reply



from math import ceil
from datetime import datetime
import csv
from pathlib import Path
import io
import time


class Logger:

    def __init__(self, **kwargs):

        self.messages = kwargs.get('messages')
        self.connection = kwargs.get('connection')
        self.commands = None

        self._logging = False

        self._writing_packet = False
        self._writing_ascii = False
        self._writing_condition = False

        self._path = str(Path.cwd()) + '/logs'
        self.packet_file = None
        self.ascii_file = None
        self.condition_file = None

        self.packet_writer = None
        self.ascii_writer = None
        self.condition_writer = None

        self.cp_nc = None

    def _get_field_names_packet(self):
        """function to return the fieldnames for the UNS logging"""

        header_fields = ['sizeHeader', 'id', 'family', 'sizeData', 'size', 'dataCheckSum', 'headerCheckSum']

        common_fields = ['version', 'offsetOfData', 'flags.posixTime', 'timeStamp', 'microSeconds']

        driver_fields = ['timestampPython']

        is_valid_field = ['isValid']

        ahrs_fields = ['serialNumber', 'operationMode',
                       'ahrsData.roll', 'ahrsData.pitch', 'ahrsData.heading',
                       'ahrsData.quaternionW', 'ahrsData.quaternionX', 'ahrsData.quaternionY', 'ahrsData.quaternionZ',
                       'ahrsData.rotationMatrix_0', 'ahrsData.rotationMatrix_1', 'ahrsData.rotationMatrix_2',
                       'ahrsData.rotationMatrix_3', 'ahrsData.rotationMatrix_4', 'ahrsData.rotationMatrix_5',
                       'ahrsData.rotationMatrix_6', 'ahrsData.rotationMatrix_7', 'ahrsData.rotationMatrix_8',
                       'declination', 'depth', 'fomAhrs', 'fomFc1']

        ins_fields = ['deltaQuaternionW', 'deltaQuaternionX', 'deltaQuaternionY', 'deltaQuaternionZ',
                      'courseOverGround', 'temperature', 'pressure',
                      'altitude', 'latitude', 'longitude', 'height',
                      'positionFrameX', 'positionFrameY', 'positionFrameZ',
                      'deltaPositionFrameX', 'deltaPositionFrameY', 'deltaPositionFrameZ',
                      'deltaPositionNucleusX', 'deltaPositionNucleusY', 'deltaPositionNucleusZ',
                      'velocityNedX', 'velocityNedY', 'velocityNedZ',
                      'velocityNucleusX', 'velocityNucleusY', 'velocityNucleusZ',
                      'deltaVelocityNedX', 'deltaVelocityNedY', 'deltaVelocityNedZ',
                      'deltaVelocityNucleusX', 'deltaVelocityNucleusY', 'deltaVelocityNucleusZ',
                      'speedOverGround', 'turnRateX', 'turnRateY', 'turnRateZ']

        imu_fields = ['status.isValid',
                      'status.hasDataPathOverrun', 'status.hasFlashUpdateFailure', 'status.hasSpiComError',
                      'status.hasLowVoltage', 'status.hasSensorFailure', 'status.hasMemoryFailure',
                      'status.hasGyro1Failure', 'status.hasGyro2Failure', 'status.hasAccelerometerFailure',
                      'accelerometer.x', 'accelerometer.y', 'accelerometer.z',
                      'gyro.x', 'gyro.y', 'gyro.z',
                      'temperature']

        mag_fields = ['status.isCompensatedForHardIron', 'status.dvlActive', 'status.dvlAcousticsActive', 'status.dvlTransmitterActive',
                      'magnetometer.x', 'magnetometer.y', 'magnetometer.z']

        dvl_fields = ['status.beam1VelocityValid', 'status.beam2VelocityValid', 'status.beam3VelocityValid',
                      'status.beam1DistanceValid', 'status.beam2DistanceValid', 'status.beam3DistanceValid',
                      'status.beam1FomValid', 'status.beam2FomValid', 'status.beam3FomValid',
                      'status.xVelocityValid', 'status.yVelocityValid', 'status.zVelocityValid',
                      'status.xFomValid', 'status.yFomValid', 'status.zFomValid',
                      'serialNumber', 'soundSpeed', 'temperature', 'pressure',
                      'velocityBeam1', 'velocityBeam2', 'velocityBeam3',
                      'distanceBeam1', 'distanceBeam2', 'distanceBeam3',
                      'fomBeam1', 'fomBeam2', 'fomBeam3', 'dtBeam1', 'dtBeam2', 'dtBeam3',
                      'timeVelBeam1', 'timeVelBeam2', 'timeVelBeam3',
                      'velocityX', 'velocityY', 'velocityZ', 'fomX', 'fomY', 'fomZ',
                      'dtXYZ', 'timeVelXYZ']

        alti_fields = ['status.altimeterDistanceValid', 'status.altimeterQualityValid', 'status.pressureValid', 'status.temperatureValid',
                       'serialNumber', 'soundSpeed', 'temperature', 'pressure',
                       'altimeterDistance', 'altimeterQuality']

        cp_fields = ['serialNumber', 'soundVelocity', 'temperature', 'pressure',
                     'cellSize', 'blanking', 'numberOfCells', 'ambiguityVelocity']

        if self.cp_nc is not None:
            for index in range(self.cp_nc * 3):
                cp_fields.append('velocityData_{}'.format(index))

            for index in range(self.cp_nc * 3):
                cp_fields.append('amplitudeData_{}'.format(index))

            for index in range(self.cp_nc * 3):
                cp_fields.append('correlationData_{}'.format(index))

        fc_fields = ['status.pointsUsedInEstimation',
                     'hardIron.x', 'hardIron.y', 'hardIron.z',
                     'sAxis_0', 'sAxis_1', 'sAxis_2',
                     'sAxis_3', 'sAxis_4', 'sAxis_5',
                     'sAxis_6', 'sAxis_7', 'sAxis_8',
                     'newPoint.x', 'newPoint.y', 'newPoint.z',
                     'fomFieldCalibration', 'coverage']

        string_fields = ['string']

        data_fields = [header_fields,
                       common_fields,
                       driver_fields,
                       is_valid_field,  # TODO: This is a temporary fix where this value needs to be in the log file, but nothing is logged to this column
                       ahrs_fields,
                       ins_fields,
                       imu_fields,
                       mag_fields,
                       dvl_fields,
                       alti_fields,
                       cp_fields,
                       fc_fields,
                       string_fields]

        field_names = list()
        for fields in data_fields:
            for field in fields:
                if field not in field_names:
                    field_names.append(field)

        return field_names

    @staticmethod
    def _get_field_names_condition():
        """function to return the fieldnames for the condition logging"""
        field_names = ['timestamp_python', 'error_message', 'failed_packet']

        return field_names

    @staticmethod
    def _get_field_names_ascii():
        """function to return the fieldnames for the info logging"""
        field_names = ['timestamp_python', 'message']

        return field_names

    def get_cp_nc(self):

        reply = self.commands.get_cur_prof(profile_range=True, cs=True, bd=True, ds=True)

        if len(reply) != 2:
            self.messages.write_warning('Unable to obtain current profile configuration')

        else:
            try:
                cp_data = reply[0].split(b',')
                cp_range = float(cp_data[0].decode())
                cp_cs = float(cp_data[1].decode())
                cp_bd = float(cp_data[2].decode())
                cp_ds = cp_data[3].rstrip(b'\r\n').decode()

                if cp_ds == '"OFF"':
                    self.cp_nc = 0

                else:
                    nc = ceil((cp_range - cp_bd) / cp_cs)
                    if nc > 0:
                        self.cp_nc = nc
                    else:
                        self.cp_nc = None

            except Exception as e:
                self.messages.write_warning('Error occured when trying to extract current profile data: {}'.format(e))

    def set_path(self, path: str):

        self._path = path.rstrip('/')

    def start(self, _converting=False) -> str:

        folder = self._path + '/' + datetime.now().strftime('%y%m%d_%H%M%S')

        Path(folder).mkdir(parents=True, exist_ok=True)

        if not _converting:
            self.messages.write_message('Logging started. Path: {}'.format(folder))
        else:
            self.messages.write_message('Converting started. Path: {}'.format(folder))

        self.packet_file = open(folder + '/nucleus_log.csv', 'w', newline='')
        self.condition_file = open(folder + '/condition_log.csv', 'w', newline='')
        self.ascii_file = open(folder + '/ascii_log.csv', 'w', newline='')

        if self.connection.get_connection_status():
            self.connection.get_info()

        if self.connection.get_all is not None and _converting is False:
            with open(folder + '/get_all.txt', 'w') as file:
                file.writelines(self.connection.get_all)

        self.packet_writer = csv.DictWriter(self.packet_file, fieldnames=self._get_field_names_packet())
        self.condition_writer = csv.DictWriter(self.condition_file, fieldnames=self._get_field_names_condition())
        self.ascii_writer = csv.DictWriter(self.ascii_file, fieldnames=self._get_field_names_ascii())

        self.packet_writer.writeheader()
        self.condition_writer.writeheader()
        self.ascii_writer.writeheader()

        self._logging = True

        return folder

    def stop(self):

        self._logging = False

        # give the writers up to 0.1 sec to complete writing
        for i in range(10):

            if not self._writing_packet and not self._writing_ascii and not self._writing_condition:
                break

            time.sleep(0.01)

        if isinstance(self.packet_file, io.TextIOWrapper):
            self.packet_file.close()

        if isinstance(self.condition_file, io.TextIOWrapper):
            self.condition_file.close()

        if isinstance(self.ascii_file, io.TextIOWrapper):
            self.ascii_file.close()
